:- module(dia,
	  [
	   load_diagram/1,    % +File
	   load_diagrams/1,   % +ListOfFiles
	   close_diagram/1,   % +File
	   close_diagrams/1,  % +ListOfFiles
	   close_all_diagrams/0,
	   import_diagram/1,  % File

	   write_facts/2, % File ListOfOptions
	   write_facts/1, % File
	   write_facts/0,

	   assert_facts/2, % ModuleIdentifier ListOfOptions
	   assert_facts/1, % ModuleIdentifier
	   assert_facts/0,
	   erase_facts/1,  % ModuleIdentifier
	   copy_facts/2,   % ModuleIdentifier(s) ModuleIdentifier

	   import_facts/0,
	   import_facts/1, % ModuleIdentifier

	   clauses/1, % -List
	   predicates_list/1 % -List
	  ]
	 ).

/**  <module> A prolog-interface to dia-uml-files

This module provides predicates to load .dia files containing
UML diagrams and make queries against it and/or save the data
contained in the form of prolog-source files.

Example
==
  load_diagram('Diagram.dia'). % Load the diagram
  import_facts.                % Tell prolog to work directly on the
			       % data - see import_facts/1.
  class_name(X, 'Class').      %  A Query
==
Example using a module:
==
load_diagram('Diagram.dia').     % Load the diagram
assert_facts(diagram, [export]). % Assert all facts into module
			         % 'diagram' and mark them as exports.
diagram:association(A).          % A Query to the asserted data.
import_facts(diagram). % Delegate queries to the diagram-module
class_operation(Class, Op) % Queries can now be made without prefix
==
Example writing to a file
==
load_diagrams(['diagram1.dia','diagram2.dia']). % Load two diagrams
write_facts('d.pl').                % Write all facts to d.pl
write_facts('mod.pl',[module, export]). % Like above but
                                        % generates a module and exports
                                        % the facts
==


@author Heiko Lewin
@license GPL
@tbd * A package_contains predicate
* Constraints on relationships
* Compressed diagrams


* NOTES
    * This works with official dia-releases (at least with 0.97.2 as
    avaiable from http://live.gnome.org/Dia ) but contains some query
    predicates official dia-releases have no diagram elements (yet) that
    correspond to them (branch-names, association-stereotypes and the
    containment-relation ). These are defined here as those are more or
    less part of the UML-standard and likely to be in one of the next
    dia-releases.
    * Though exporting any rules or facts for querying
    diagram data directly may seem clumsy this way clauses holding the
    diagram data can be merged into the user-module from different source
    modules and be worked with there without name-clashes.
    * This does not really follow the UML- or XMI-standards in how
    diagram data is represented but is close to the layout of dia-files.
    * This currently only works with SWI-Prolog without further
    extensions when it comes down to free software. GNU-Prolog is
    missing a xml-parser and the code_type/2 relation. This has not
    been tested with commerical Prolog-systems.

*/

:- use_module(library(sgml)).
:- dynamic(model/2).

%%	predicates_list( -List )
%
%	Unified List with a list of Predicate/Arity tuples that
%	define the predicates that are used to make queries
%	against the diagram-data.
%
%	The predicate names should be self-explanatory.
%	The predicates that take one argument are used to query for
%	existence and unify their argument with a term representing
%	the model-element, all other predicates unify the first argument
%	with such a term and the second with the thing asked for.
%
%	* existential_query(?ElementOrHandle)
%	* query_for_an_attribute(?ElementOrHandle, ?VariableOrTerm)
%
%	NOTE: The first argument can either be an atomic identifier
%	( in cases facts got asserted ) or an xml-element if
%	working directly on the xml-file, so best is to look
%	at it as black box.
predicates_list(
    [
    association/1,
    association_direction/2,
    association_id/2,
    association_name/2,
    association_from_arrow/2,
    association_from_id/2,
    association_from_multiplicity/2,
    association_from_role/2,
    association_from_visibility/2,
    association_stereotype/2,
    association_text/2,
    association_to_arrow/2,
    association_to_id/2,
    association_to_multiplicity/2,
    association_to_role/2,
    association_to_visibility/2,
    association_type/2,

    attribute/1,
    attribute_abstract/2,
    attribute_comment/2,
    attribute_id/2,
    attribute_name/2,
    attribute_static/2,
    attribute_type/2,
    attribute_value/2,
    attribute_visibility/2,
    attribute_inherited/2,

    branch/1,
    branch_id/2,
    branch_name/2,
    branch_stereotype/2,

    constraint/1,
    constraint_id/2,
    constraint_what/2,
    constraint_expression/2,

    class/1,
    class_abstract/2,
    class_attribute/2,
    class_attributes/2,
    class_comment/2,
    class_id/2,
    class_name/2,
    class_operation/2,
    class_operations/2,
    class_stereotype/2,
    class_template/2,
    class_template_parameter/2,
    class_template_parameters/2,

    operation/1,
    operation_abstract/2,
    operation_comment/2,
    operation_id/2,
    operation_inheritance/2,
    operation_name/2,
    operation_parameter/2,
    operation_parameters/2,
    operation_static/2,
    operation_stereotype/2,
    operation_type/2,
    operation_visibility/2,
    operation_query/2,

    package/1,
    package_id/2,
    package_name/2,
    package_stereotype/2,

    parameter/1,
    parameter_comment/2,
    parameter_direction/2,
    parameter_id/2,
    parameter_name/2,
    parameter_type/2,
    parameter_value/2,

    template_parameter/1,
    template_parameter_id/2,
    template_parameter_name/2,
    template_parameter_type/2

    ]).

element_type(element(Tag, _, _), Tag).
element_attributes(element(_,Attributes,_), Attributes).
element_attribute(Element, Attribute) :-
	element_attributes(Element, Attributes),
	member(Attribute, Attributes),
	Attribute = (_=_).
element_content(element(_,_,Content), Content).
element_element(Element, ContainedElement) :-
	ContainedElement = element(_,_,_),
	element_content(Element, Content),
	member(ContainedElement, Content).

filespec(FileSpec, Extension, Base, AbsoluteFilename) :-
	nonvar(FileSpec),
	nonvar(Extension),
	file_base_name(FileSpec, FilePart),
	(   atom_concat(Base, Extension, FilePart)
	->  Filename = FileSpec
	;   Base = FilePart,
	    atom_concat(FileSpec, Extension, Filename)),
	absolute_file_name(Filename, AbsoluteFilename).

%%	load_diagrams( +ListOfFiles : list )
%
%	Calls load_diagram/1 for each file in ListOfFiles.
load_diagrams([]).
load_diagrams([M|L]) :- load_diagram(M), load_diagrams(L).

%%	load_diagram( +FileSpec )
%
%	Loads the diagram contents of FileSpec which has to be
%	an uncompressed .dia file and asserts the raw data
%	in the dia-module.
load_diagram(FileSpec) :-
	filespec(FileSpec, '.dia', _, File),
	load_xml_file(File, [M]),
	retractall(model(File,_)),
	assert(model(File,M)).

%%	close_diagram( +FileSpec )
%
%	Retracts the diagram data loaded from FileSpec from the
%	dia-module.
close_diagram(N) :-
	filespec(N, '.dia', _, File),
	retractall(model(File, _)).

%%	close_diagrams( +ListOfFiles )
%
%	Calls close_diagram/1 for each member of ListOfFiles.
close_diagrams(L) :- foreach(member(X, L), close_diagram(X)).

%%	close_all_diagrams
%
%	Retracts all diagram-data from the dia-module.
close_all_diagrams :- retractall(model(_,_)).

model(M) :- model(_,M).

model_id(M, I) :-
	model(_, M),
	term_hash(M, I).

dia_string(NameElement, Name) :-
	element_element(NameElement, StringElement),
	element_type(StringElement, 'dia:string'),
	element_content(StringElement, [DecName]),
/*
	append("#", NameCodes, HlfDecName),
	append(HlfDecName, "#", DecName),
	atom_codes(Name, NameCodes). */

	atom_concat('#', HlfDecName, DecName),
	atom_concat(Name, '#', HlfDecName).


element_string(Elem, String) :- dia_string(Elem, String).
element_boolean(Elem, Boolean) :-
	element_element(Elem, ValElem),
	element_type(ValElem, 'dia:boolean'),
	element_attribute(ValElem, val=Boolean).
element_enum(Elem, Enum) :-
	element_element(Elem, ValElem),
	element_type(ValElem, 'dia:enum'),
	element_attribute(ValElem, val=Enum).

element_string_attribute(Elem, AttrName, String) :-
	element_element(Elem, Elem2),
	element_type(Elem2, 'dia:attribute'),
	element_attribute(Elem2, name=AttrName),
	element_string(Elem2, String).
element_boolean_attribute(Elem, AttrName, Boolean) :-
	element_element(Elem, Elem2),
	element_type(Elem2, 'dia:attribute'),
	element_attribute(Elem2, name=AttrName),
	element_boolean(Elem2, Boolean).
element_enum_attribute(Elem, AttrName, Enum) :-
	element_element(Elem, Elem2),
	element_type(Elem2, 'dia:attribute'),
	element_attribute(Elem2, name=AttrName),
	element_enum(Elem2, Enum).


dia_visibility('0', public).
dia_visibility('1', private).
dia_visibility('2', protected).
dia_visibility('3', implementation).

dia_inheritance('2', final).
dia_inheritance('1', virtual).
dia_inheritance('0', abstract).

dia_parameter_kind('0', undefined).
dia_parameter_kind('1', in).
dia_parameter_kind('2', out).
dia_parameter_kind('3', inout).

dia_direction('0', undefined).
dia_direction('1', a_to_b).
dia_direction('2', b_to_a).

dia_assoc_type('0', association).
dia_assoc_type('1', aggregation).
dia_assoc_type('2', composition).


element_hash(Element, Hash) :- term_hash(Element, Hash).

dia_layer(Diagram, LayerElement) :-
	model(Diagram),
	element_element(Diagram, LayerElement),
	element_type(LayerElement, 'dia:layer').

dia_layer(L) :- dia_layer(_,L).

class(Model, C) :-
	dia_layer(Model, L),
	element_element(L, C),
	element_type(C, 'dia:object'),
	element_attribute(C, type='UML - Class').

class(C) :- class(_, C).

class_id(Class, Id) :-
	class(Model, Class),
	element_attribute(Class, id=ClassId),
	model_id(Model, ModelId),
	atom_concat(ModelId, ClassId, Id).

class_name(Class, Name) :-
	class(Class),
	element_string_attribute(Class, name, Name).

class_stereotype(Class, Stereo) :-
	class(Class),
	element_string_attribute(Class, stereotype, Stereo).

class_comment(Class, Comment) :-
	class(Class),
	element_string_attribute(Class, comment, Comment).

class_abstract(Class, Abstract) :-
	class(Class),
	element_boolean_attribute(Class, abstract, Abstract).

class_template(Class, Abstract) :-
	class(Class),
	element_boolean_attribute(Class, template, Abstract).

class_template_parameters(Class, Abstract) :-
	class(Class),
	element_element(Class, Abstract),
	element_type(Abstract, 'dia:attribute'),
	element_attribute(Abstract, name=templates).

class_template_parameter(Class, Abstract) :-
	class_template_parameters(Class, P),
	element_element(P, Abstract),
	element_type(Abstract, 'dia:composite'),
	element_attribute(Abstract, type=umlformalparameter).

class_attributes(Class, Attributes) :-
	class(Class),
	element_element(Class, Attributes),
	element_type(Attributes, 'dia:attribute'),
	element_attribute(Attributes, name=attributes).

class_attribute(Class, Attribute) :-
	class_attributes(Class, Attributes),
	element_element(Attributes, Attribute),
	element_type(Attribute, 'dia:composite'),
	element_attribute(Attribute, type=umlattribute).

class_operations(Class, Operations) :-
	class(Class),
	element_element(Class, Operations),
	element_type(Operations, 'dia:attribute'),
	element_attribute(Operations, name=operations).

class_operation(Class, Operation) :-
	class_operations(Class, Operations),
	element_element(Operations, Operation),
	element_type(Operation, 'dia:composite'),
	element_attribute(Operation, type=umloperation).

template_parameter(P) :- class_template_parameter(_,P).

template_parameter_id(A, Id) :-
	class_template_parameter(Class, A),
	class_id(Class,ClassId),
	term_hash(A, H),
	atom_number(Ha, H),
	atom_concat(ClassId, 'T', X),
	atom_concat(X, Ha, Id).

template_parameter_name(TP, N) :-
	template_parameter(TP),
	element_string_attribute(TP, name, N).

template_parameter_type(TP, N) :-
	template_parameter(TP),
	element_string_attribute(TP, type, N).

attribute(A) :-
	class(Class),
	class_attribute(Class, A).

attribute_id(A, Id) :-
	class_attribute(Class, A),
	class_id(Class,ClassId),
	term_hash(A, H),
	atom_number(Ha, H),
	atom_concat(ClassId, 'A', X),
	atom_concat(X, Ha, Id).

attribute_name(Attribute, Name, IsInherited) :-
	class(Class),
	class_attribute(Class, Attribute),
	element_string_attribute(Attribute, name, FullName),
	(   atom_concat('/', Name, FullName)
	->  IsInherited = true
	;   Name = FullName,
	    IsInherited = false).

attribute_inherited(Attribute, IsInherited) :-
	attribute_name(Attribute, _, IsInherited).

attribute_name(Attribute, Name) :- attribute_name(Attribute, Name, _).

attribute_type(Attribute, Name) :-
	attribute(Attribute),
	element_string_attribute(Attribute, type, AName),
	(   AName == ''
	->  var(Name)
	;   Name = AName).

attribute_value(Attribute, Name) :-
	attribute(Attribute),
	element_string_attribute(Attribute, value, AName),
	(   AName == ''
	->  var(Name)
	;   Name = '').

attribute_comment(Attribute, Name) :-
	attribute(Attribute),
	element_string_attribute(Attribute, comment, Name).

attribute_visibility(Attribute, Name) :-
	attribute(Attribute),
	element_enum_attribute(Attribute, visibility, X),
	dia_visibility(X,Name).

attribute_static(Attribute, Static) :-
	attribute(Attribute),
	element_boolean_attribute(Attribute, class_scope, Static).

attribute_abstract(Attribute, Static) :-
	attribute(Attribute),
	element_boolean_attribute(Attribute, abstract, Static).

operation(Operation) :-
	class_operation(_, Operation).

operation_id(A, Id) :-
	class_operation(Class, A),
	class_id(Class,ClassId),
	term_hash(A, H),
	atom_number(Ha, H),
	atom_concat(ClassId, 'O', X),
	atom_concat(X, Ha, Id).

operation_name(Attribute, Name) :-
	operation(Attribute),
	element_string_attribute(Attribute, name, Name).

operation_stereotype(Attribute, Name) :-
	operation(Attribute),
	element_string_attribute(Attribute, stereotype, Name).

operation_type(Attribute, Name) :-
	operation(Attribute),
	element_string_attribute(Attribute, type, AName),
	(   AName == ''
	->  var(Name)
	;   Name = AName).


operation_comment(Attribute, Name) :-
	operation(Attribute),
	element_string_attribute(Attribute, comment, Name).

operation_visibility(Attribute, Name) :-
	operation(Attribute),
	element_enum_attribute(Attribute, visibility, C),
	dia_visibility(C, Name).

operation_static(Attribute, Static) :-
	operation(Attribute),
	element_boolean_attribute(Attribute, class_scope, Static).

operation_abstract(Attribute, Static) :-
	operation(Attribute),
	element_boolean_attribute(Attribute, abstract, Static).

operation_query(Attribute, Static) :-
	operation(Attribute),
	element_boolean_attribute(Attribute, query, Static).

operation_inheritance(Attribute, Name) :-
	operation(Attribute),
	element_enum_attribute(Attribute, inheritance_type, X),
	dia_inheritance(X, Name).

operation_parameters(Operation, Parameters) :-
	operation(Operation),
	element_element(Operation, Parameters),
	element_type(Parameters, 'dia:attribute'),
	element_attribute(Parameters, name=parameters).

operation_parameter(Operation, Parameter) :-
	operation_parameters(Operation, P),
	element_element(P, Parameter),
	element_type(Parameter, 'dia:composite'),
	element_attribute(Parameter, type=umlparameter).

parameter(P) :- operation_parameter(_, P).

parameter_id(P, I) :-
	operation_parameter(O, P),
	operation_id(O,OpId),
	term_hash(P, H),
	atom_number(Ha, H),
	atom_concat(OpId, 'P', X),
	atom_concat(X, Ha, I).

parameter_name(Attribute, Name) :-
	parameter(Attribute),
	element_string_attribute(Attribute, name, Name).

parameter_type(Attribute, Name) :-
	parameter(Attribute),
	element_string_attribute(Attribute, type, AName),
	(   AName = ''
	->  var(Name)
	;   Name = AName).


parameter_value(Attribute, Name) :-
	parameter(Attribute),
	element_string_attribute(Attribute, value, AName),
	(   AName = ''
	->  var(Name)
	;   Name = AName).

parameter_comment(Attribute, Name) :-
	parameter(Attribute),
	element_string_attribute(Attribute, comment, Name).

parameter_direction(Parameter, Direction) :-
	parameter(Parameter),
	element_enum_attribute(Parameter, kind, X),
	dia_parameter_kind(X, Direction).

association(Model, A) :-
	dia_layer(Model, L),
	element_element(L, A),
	element_type(A, 'dia:object'),
	element_attribute(A, type=Type),
	(   Type='UML - Association'
	;   Type='UML - Generalization'
	;   Type='UML - Realizes'
	;   Type='UML - Dependency'
	;   Type='UML - Implements'
	;   Type='UML - Contains').

association(A) :- association(_,A).

association_id(A, ID) :-
	association(Model, A),
	element_attribute(A, id=AssocID),
	model_id(Model, MID),
	atom_concat(MID,AssocID, ID).

association_name(A, N) :-
	association(A),
	element_string_attribute(A, name, N).

association_stereotype(A, N) :-
	association(A),
	(   element_string_attribute(A, stereotype, N)
	->  true
	;   N = '').

association_text(A, T) :-
	association(A),
	(   element_string_attribute(A, text, T)
	->  true
	;   T = '').

association_direction(A, N) :-
	association(A),
	(   element_enum_attribute(A, direction, X)
	->  dia_direction(X, N)
	;   N = undefined).

association_type(A, N) :-
	association(A),
	element_attribute(A, type=T),
	(   T =='UML - Generalization'
	->  N = generalization
	;   T == 'UML - Realizes'
	->  N = realization
	;   T == 'UML - Dependency'
	->  N = dependency
	;   T == 'UML - Contains'
	->  N = containment
	;   T == 'UML - Implements'
	->  N = implementation
	;   T == 'UML - Association'
	->  element_enum_attribute(A, assoc_type, X),
	    dia_assoc_type(X, N)).

association_role_a(A, R) :-
	association(A),
	(   element_string_attribute(A, role_a, R)
	->  true
	;   R='').

association_multiplicity_a(A, R) :-
	association(A),
	(   element_string_attribute(A, multipicity_a, R)
	->  true
	;   element_string_attribute(A, multiplicity_a, R)
	->  true
	;   R='').

association_visibility_a(A, R) :-
	association(A),
	(   element_enum_attribute(A, visibility_a, X)
	->  dia_visibility(X,R)
	;   R=undefined).

association_arrow_a(A, R) :-
	association(A),
	(   element_boolean_attribute(A, show_arrow_a, R)
	->  true
	;   R=undefined).

association_role_b(A, R) :-
	association(A),
	(   element_string_attribute(A, role_b, R)
	->  true
	;   R='').

association_multiplicity_b(A, R) :-
	association(A),
	(   element_string_attribute(A, multipicity_b, R)
	->  true
	;   element_string_attribute(A, multiplicity_b, R)
	->  true
	;   R='').

association_visibility_b(A, R) :-
	association(A),
	(   element_enum_attribute(A, visibility_b, X)
	->  dia_visibility(X,R)
	;   R=undefined).

association_arrow_b(A, R) :-
	association(A),
	(   element_boolean_attribute(A, show_arrow_b, R)
	->  true
	;   R=undefined).

association_from_id(A, X) :-
	association_direction(A, D),
	(   D = b_to_a
	->  association_end_b(A, X)
	;   association_end_a(A, X)).

association_from_role(A, X) :-
	association_direction(A, D),
	(   D = b_to_a
	->  association_role_b(A, X)
	;   association_role_a(A, X)).

association_from_multiplicity(A, X) :-
	association_direction(A, D),
	(   D = b_to_a
	->  association_multiplicity_b(A, X)
	;   association_multiplicity_a(A, X)).

association_from_arrow(A, X) :-
	association_direction(A, D),
	(   D = b_to_a
	->  association_arrow_b(A, X)
	;   association_arrow_a(A, X)).

association_from_visibility(A, X) :-
	association_direction(A, D),
	(   D = b_to_a
	->  association_visibility_b(A, X)
	;   association_visibility_a(A, X)).

association_to_id(A, X) :-
	association_direction(A, D),
	(   D = b_to_a
	->  association_end_a(A, X)
	;   association_end_b(A, X)).

association_to_role(A, X) :-
	association_direction(A, D),
	(   D = b_to_a
	->  association_role_a(A, X)
	;   association_role_b(A, X)).

association_to_multiplicity(A, X) :-
	association_direction(A, D),
	(   D = b_to_a
	->  association_multiplicity_a(A, X)
	;   association_multiplicity_b(A, X)).

association_to_arrow(A, X) :-
	association_direction(A, D),
	(   D = b_to_a
	->  association_arrow_a(A, X)
	;   association_arrow_b(A, X)).

association_to_visibility(A, X) :-
	association_direction(A, D),
	(   D = b_to_a
	->  association_visibility_a(A, X)
	;   association_visibility_b(A, X)).


association_connections(A, C) :-
	association(A),
	element_element(A, C),
	element_type(C, 'dia:connections').

association_end_a(A, CA) :-
	association_connections(A, C),
	element_element(C, X),
	element_type(X, 'dia:connection'),
	element_attribute(X, handle='0'),
	element_attribute(X, to=Id),
	association(Model, A),
	model_id(Model, ModelId),
	atom_concat(ModelId, Id, CA).

association_end_b(A, CA) :-
	association_connections(A, C),
	element_element(C, X),
	element_type(X, 'dia:connection'),
	element_attribute(X, handle='1'),
	element_attribute(X, to=Id),
	association(Model, A),
	model_id(Model, ModelId),
	atom_concat(ModelId, Id, CA).

large_package(M, X) :-
	dia_layer(M, L),
	element_element(L, X),
	element_type(X, 'dia:object'),
	element_attribute(X, type='UML - LargePackage').

large_package(X) :- large_package(_,X).
large_package_id(X, Id) :-
	large_package(M, X),
	element_attribute(X, id=PackageId),
	model_id(M, ModelId),
	atom_concat(ModelId, PackageId, Id).
large_package_name(X, N) :-
	large_package(X),
	element_string_attribute(X, name, N).
large_package_stereotype(X, N) :-
	large_package(X),
	element_string_attribute(X, stereotype, N).

small_package(M, X) :-
	dia_layer(M, L),
	element_element(L, X),
	element_type(X, 'dia:object'),
	element_attribute(X, type='UML - SmallPackage').

small_package(X) :- small_package(_,X).
small_package_id(X, Id) :-
	small_package(M, X),
	element_attribute(X, id=PackageId),
	model_id(M, ModelId),
	atom_concat(ModelId, PackageId, Id).
small_package_text(X, N) :-
	small_package(X),
	element_element(X, Xn),
	element_attribute(Xn, name=text),
	element_type(Xn, 'dia:attribute'),
	element_element(Xn, Xnn),
	element_type(Xnn, 'dia:composite'),
	element_attribute(Xnn, type=text),
	element_string_attribute(Xnn, string, N).

small_package_stereotype(X, N) :-
	small_package(X),
	element_string_attribute(X, stereotype, N).

branch(M, B) :-
	dia_layer(M, L),
	element_element(L, B),
	element_type(B, 'dia:object'),
	element_attribute(B, type='UML - Branch').

branch(B) :- branch(_, B).

branch_id(B, I) :-
	branch(M, B),
	element_attribute(B, id=BranchId),
	model_id(M, ModelId),
	atom_concat(ModelId, BranchId, I).

branch_name(B, N) :-
	branch(B),
	(   element_string_attribute(B, name, N)
	->  true
	;   N='').

branch_stereotype(B, N) :-
	branch(B),
	(   element_string_attribute(B, stereotype, N)
	->  true
	;   N='').


package(P) :- small_package(P) ; large_package(P).
package_id(P, Id) :- small_package_id(P, Id) ; large_package_id(P, Id).
package_name(P, Id) :- large_package_name(P, Id); small_package_text(P, Id).
package_stereotype(P, S) :-
	large_package_stereotype(P, S) ; small_package_stereotype(P, S).

constraint(Model, C) :-
	dia_layer(Model, L),
	element_element(L, C),
	element_type(C, 'dia:object'),
	element_attribute(C, type='UML - Constraint').
constraint(C) :- constraint(_, C).
constraint_id(C, Id) :-
	constraint(M, C),
	model_id(M, Mid),
	element_attribute(C, id=Cid),
	atom_concat(Mid, Cid, Id).
constraint_what(C, Id) :-
	constraint(M, C),
	model_id(M, Mid),
	element_element(C, Cs),
	element_type(Cs, 'dia:connections'),
	element_element(Cs, Con),
	element_type(Con, 'dia:connection'),
	element_attribute(Con, handle='1'),
	element_attribute(Con, to=Wid),
	atom_concat( Mid, Wid, Id ).

constraint_expression(C, Expr) :-
	constraint(C),
	element_string_attribute(C, constraint, Expr).

element_id(Element, Id) :-
	(   class_id(Element, Id)
	;   operation_id(Element, Id)
	;   attribute_id(Element, Id)
	;   parameter_id(Element, Id)
	;   association_id(Element, Id)
	;   template_parameter_id(Element, Id)
	;   branch_id(Element, Id)
	;   large_package_id(Element, Id)
	;   small_package_id(Element, Id)
	;   constraint_id(Element, Id)).

clause(C) :-
	predicates_list(Predicates),
	member(Predicate/Arity, Predicates),
	functor(Pred, Predicate, Arity),
	(   Arity =:= 1,
	    Pred =.. [ Predicate, Argument ],
	    call(Pred),
	    atom_concat(Predicate, '_id', IdPredicate),
	    Query =.. [ IdPredicate, Argument, Id ],
	    call(Query),
	    C =.. [ Predicate, Id ]
	;   Arity =:= 2,
	    Pred =.. [ Predicate, Object, Value ],
	    call(Pred),
	    element_id(Object, ObjectId),
	    (	(   Predicate == class_attributes
		;   Predicate == class_operations
		;   Predicate == operation_parameters
		;   Predicate == class_template_parameters)
	    ->	findall(Id, ( element_element(Value, E), element_id(E,Id)), Ids),
		C =.. [ Predicate, ObjectId, Ids ]
	    ;	(   Predicate == class_attribute
		;   Predicate == class_operation
		;   Predicate == class_template_parameter
		;   Predicate == operation_parameter)
	    ->	element_id(Value, Id),
		C =.. [ Predicate, ObjectId, Id ]
	    ;	C =.. [ Predicate, ObjectId, Value])
	;   C = (Pred :- fail)).

clauses(Clauses, Specs) :-
	findall( C, clause(C), Clauses ),
	clauses_specs(Clauses, Specs).

clauses_specs(Clauses, Specs) :-
	findall(Pn/Pa,
		(   member(C, Clauses),
		    functor(C, Pn, Pa)),
		List),
	list_to_set(List, Specs).

%%	clauses(-List)
%
%	Unifies List with a list of all goals in predicates_list/1
%	that succeed in the dia-module. The list then contains
%	all relevant diagram data in the form of prolog-clauses
%	and can be written to files or asserted into
%       appropriate modules (see write_facts/2 and assert_facts/2).
clauses(Clauses) :- clauses(Clauses, _).

%%	write_facts( +FileSpec, +Options )
%
%	Calls clauses/1 and writes the clauses representing
%	the diagram-data into the file specified by FileSpec.
%	Options is a list of options:
%	* module
%	Generates a module/2 directive.
%	* dynamic
%	Declares the clauses written as dynamic (see dynamic/1).
%	* export
%	Marks the clauses written as exports (see export/1).
%	* multifile
%	Declares the clauses written as multifile (see multifile/1).
write_facts(Dest, Opts) :-
	telling(X),
	filespec(Dest, '.pl', Name, File),
	tell(File),
	(   memberchk(module, Opts)
	->  writeq( :- module(Name, []) ), write('.'), nl
	;   true),
	(   memberchk(dynamic, Opts)
	->  predicates_list(Predicates),
	    foreach(member(Predicate, Predicates),
		    (	writeq( :- dynamic(Predicate) ), writeln('.') ))
	;   true),
	(   memberchk(multifile, Opts)
	->  predicates_list(Predicates),
	    foreach(member(Predicate, Predicates),
		    (	writeq( :- multifile(Predicate) ), writeln('.') ))
	;   true),
	(   memberchk(export, Opts)
	->  predicates_list(Predicates),
	    foreach(member(Predicate, Predicates),
		    (	writeq( :- export(Predicate) ), writeln('.') ))
	;   true),
	write_facts,
	told,
	tell(X).

%%	write_facts( +FileSpec )
%
%	Shorthand for write_facts( FileSpec, [] ).
write_facts(File) :- write_facts(File, []).

%%	write_facts
%
%	Calls clauses/1 and prints the resulting list to the
%	current output stream using writeq/1.
write_facts :-
	clauses(Clauses),
	foreach(member(Clause, Clauses), (writeq(Clause), writeln('.'))).


%%	erase_facts(+ModuleIdentifier)
%
%	Erases all clauses matching one of the predicates
%	in predicates_list/1 from the module identified
%	by ModuleIdentifier.
erase_facts(Mod) :-
	predicates_list(L),
	foreach( (   member(Pn/Pa, L),
		     functor(P, Pn, Pa)),
		 foreach(clause(Mod:P, _, R), erase(R))).

%%	assert_facts(+ModuleIdentifier, Options)
%
%	Calls clauses/1 and asserts the clauses into the
%	module specified by ModuleIdentifier.
%	Options is a list of options:
%	* export
%	Mark the clauses asserted as exports (see export/1).
assert_facts(Mod, Opts) :-
	clauses(X),
	foreach(member(C, X), assert(Mod:C)),
	(   memberchk(export, Opts)
	->  predicates_list(Predicates),
	    foreach(member(Predicate, Predicates),
		    export(Mod:Predicate))
	;   true).

%%	assert_facts(Module)
%
%	Shorthand for assert_facts(Module, []).
assert_facts(Mod) :- assert_facts(Mod, []).

%%	assert_facts
%
%	Shorthand for assert_facts(user, []).
assert_facts :- assert_facts(user).

%%	import_facts(Module)
%
%	Asserts rules into the user module that delegate queries
%	for diagram-data to Module.
import_facts(FromModule) :-
	predicates_list(Predicates),
	foreach( (   member(Pn/Pa, Predicates),
		     functor(P, Pn, Pa)),
		 (   user:assert( P :- FromModule:P ) )).

%%	import_facts
%
%	Equivalent to import_facts(dia).
import_facts :- import_facts(dia).


erase_import_rules :-
	predicates_list(Predicates),
	foreach( (   member(Pn/Pa, Predicates),
		     functor(P, Pn, Pa)),
		 user:retractall(P)).

copy_facts_(Src, Dest) :-
	predicates_list(Pl),
	findall( P,
		 (   member(Pn/Pa, Pl),
		     functor(P, Pn, Pa),
		     call(Src:P)),
		 Facts),
	foreach(member(F, Facts), assert(Dest:F)).

%%	copy_facts(+ListOfModules, +DestinationModule)
%
%	Asserts all goals in predicates_list/1 that succeed in
%	any of the modules in ListOfModules into DestinationModule.
copy_facts(X, Dest) :-
	atom(X),
	(   X == []
	->  true
	;   copy_facts_(X, Dest)).
copy_facts([M|Modules], Dest) :-
	copy_facts(M, Dest),
	copy_facts(Modules, Dest).

%%	import_diagram( +FileSpec )
%
%	Loads the diagram specified by FileSpec and asserts the
%	facts representing the diagram-data into a module with
%	the same name.
import_diagram(FileSpec) :-
	load_diagram(FileSpec),
	filespec(FileSpec,'.dia', Name, _),
	erase_facts(Name),
	assert_facts(Name),
	close_diagram(Name).













